"""ATR class managing some of the Answer To Reset content.

__author__ = "http://www.gemalto.com"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

from __future__ import print_function
from smartcard.Exceptions import SmartcardException
from smartcard.util import toHexString


class ATR(object):
    """ATR class."""

    clockrateconversion = [372, 372, 558, 744, 1116, 1488, 1860, 'RFU',
                           'RFU', 512, 768, 1024, 1536, 2048, 'RFU', 'RFU',
                           'RFU']
    bitratefactor = ['RFU', 1, 2, 4, 8, 16, 32, 'RFU', 12, 20, 'RFU',
                     'RFU', 'RFU', 'RFU', 'RFU', 'RFU']
    currenttable = [25, 50, 100, 'RFU']

    def __init__(self, bytes):
        """Construct a new atr from bytes."""
        self.bytes = bytes
        self.__initInstance__()

    def __checksyncbyte__(self):
        """Check validity of TS."""
        if not 0x3b == self.bytes[0] and not 0x03f == self.bytes[0]:
            raise SmartcardException("invalid TS 0x%-0.2x" % self.bytes[0])

    def __initInstance__(self):
        """
        Parse ATR and initialize members:
          - TS: initial character
          - T0: format character
          - TA[n], TB[n], TC[n], TD[n], for n=0,1,...: protocol parameters

        @note: protocol parameters indices start at 0, e.g.
        TA[0], TA[1] correspond to the ISO standard TA1, TA2
        parameters

          - historicalBytes: the ATR T1, T2, ..., TK historical bytes
          - TCK: checksum byte (only for protocols different from T=0)
          - FI: clock rate conversion factor
          - DI: voltage adjustment factor
          - PI1: programming voltage factor
          - II: maximum programming current factor
          - N: extra guard time
        """
        self.__checksyncbyte__()

        # initial character
        self.TS = self.bytes[0]

        # format character
        self.T0 = self.bytes[1]

        # count of historical bytes
        self.K = self.T0 & 0x0f

        # initialize optional characters lists
        self.TA = []
        self.TB = []
        self.TC = []
        self.TD = []
        self.Y = []
        self.hasTA = []
        self.hasTB = []
        self.hasTC = []
        self.hasTD = []

        TD = self.T0
        hasTD = 1
        n = 0
        offset = 1
        self.interfaceBytesCount = 0
        while hasTD:
            self.Y += [TD >> 4 & 0x0f]

            self.hasTD += [(self.Y[n] & 0x08) != 0]
            self.hasTC += [(self.Y[n] & 0x04) != 0]
            self.hasTB += [(self.Y[n] & 0x02) != 0]
            self.hasTA += [(self.Y[n] & 0x01) != 0]

            self.TA += [None]
            self.TB += [None]
            self.TC += [None]
            self.TD += [None]

            if self.hasTA[n]:
                self.TA[n] = self.bytes[offset + self.hasTA[n]]
            if self.hasTB[n]:
                self.TB[n] = self.bytes[offset + self.hasTA[n] + self.hasTB[n]]
            if self.hasTC[n]:
                self.TC[n] = self.bytes[offset +
                                        self.hasTA[n] +
                                        self.hasTB[n] +
                                        self.hasTC[n]]
            if self.hasTD[n]:
                self.TD[n] = self.bytes[offset +
                                        self.hasTA[n] +
                                        self.hasTB[n] +
                                        self.hasTC[n] +
                                        self.hasTD[n]]

            self.interfaceBytesCount += self.hasTA[n] +\
                self.hasTB[n] +\
                self.hasTC[n] +\
                self.hasTD[n]
            TD = self.TD[n]
            hasTD = self.hasTD[n]
            offset = offset + self.hasTA[n] + self.hasTB[n] +\
                self.hasTC[n] + self.hasTD[n]
            n = n + 1

        # historical bytes
        self.historicalBytes = self.bytes[offset + 1:offset + 1 + self.K]

        # checksum
        self.hasChecksum = (len(self.bytes) == offset + 1 + self.K + 1)
        if self.hasChecksum:
            self.TCK = self.bytes[-1]
            checksum = 0
            for b in self.bytes[1:]:
                checksum = checksum ^ b
            self.checksumOK = (checksum == 0)
        else:
            self.TCK = None

        # clock-rate conversion factor
        if self.hasTA[0]:
            self.FI = self.TA[0] >> 4 & 0x0f
        else:
            self.FI = None

        # bit-rate adjustment factor
        if self.hasTA[0]:
            self.DI = self.TA[0] & 0x0f
        else:
            self.DI = None

        # maximum programming current factor
        if self.hasTB[0]:
            self.II = self.TB[0] >> 5 & 0x03
        else:
            self.II = None

        # programming voltage factor
        if self.hasTB[0]:
            self.PI1 = self.TB[0] & 0x1f
        else:
            self.PI1 = None

        # extra guard time
        self.N = self.TC[0]

    def getChecksum(self):
        """Return the checksum of the ATR. Checksum is mandatory only
        for T=1."""
        return self.TCK

    def getHistoricalBytes(self):
        """Return historical bytes."""
        return self.historicalBytes

    def getHistoricalBytesCount(self):
        """Return count of historical bytes."""
        return len(self.historicalBytes)

    def getInterfaceBytesCount(self):
        """Return count of interface bytes."""
        return self.interfaceBytesCount

    def getTA1(self):
        """Return TA1 byte."""
        return self.TA[0]

    def getTB1(self):
        """Return TB1 byte."""
        return self.TB[0]

    def getTC1(self):
        """Return TC1 byte."""
        return self.TC[0]

    def getTD1(self):
        """Return TD1 byte."""
        return self.TD[0]

    def getBitRateFactor(self):
        """Return bit rate factor."""
        if self.DI is not None:
            return ATR.bitratefactor[self.DI]
        else:
            return 1

    def getClockRateConversion(self):
        """Return clock rate conversion."""
        if self.FI is not None:
            return ATR.clockrateconversion[self.FI]
        else:
            return 372

    def getProgrammingCurrent(self):
        """Return maximum programming current."""
        if self.II is not None:
            return ATR.currenttable[self.II]
        else:
            return 50

    def getProgrammingVoltage(self):
        """Return programming voltage."""
        if self.PI1 is not None:
            return 5 * (1 + self.PI1)
        else:
            return 5

    def getGuardTime(self):
        """Return extra guard time."""
        return self.N

    def getSupportedProtocols(self):
        """Returns a dictionnary of supported protocols."""
        protocols = {}
        for td in self.TD:
            if td is not None:
                strprotocol = "T=%d" % (td & 0x0F)
                protocols[strprotocol] = True
        if not self.hasTD[0]:
            protocols['T=0'] = True
        return protocols

    def isT0Supported(self):
        """Return True if T=0 is supported."""
        protocols = self.getSupportedProtocols()
        return 'T=0' in protocols

    def isT1Supported(self):
        """Return True if T=1 is supported."""
        protocols = self.getSupportedProtocols()
        return 'T=1' in protocols

    def isT15Supported(self):
        """Return True if T=15 is supported."""
        protocols = self.getSupportedProtocols()
        return 'T=15' in protocols

    def dump(self):
        """Dump the details of an ATR."""

        for i in range(0, len(self.TA)):
            if self.TA[i] is not None:
                print("TA%d: %x" % (i + 1, self.TA[i]))
            if self.TB[i] is not None:
                print("TB%d: %x" % (i + 1, self.TB[i]))
            if self.TC[i] is not None:
                print("TC%d: %x" % (i + 1, self.TC[i]))
            if self.TD[i] is not None:
                print("TD%d: %x" % (i + 1, self.TD[i]))

        print('supported protocols ' + ','.join(self.getSupportedProtocols()))
        print('T=0 supported: ' + str(self.isT0Supported()))
        print('T=1 supported: ' + str(self.isT1Supported()))

        if self.getChecksum():
            print('checksum: %d' % self.getChecksum())

        print('\tclock rate conversion factor: ' +
              str(self.getClockRateConversion()))
        print('\tbit rate adjustment factor: ' + str(self.getBitRateFactor()))

        print('\tmaximum programming current: ' +
              str(self.getProgrammingCurrent()))
        print('\tprogramming voltage: ' + str(self.getProgrammingVoltage()))

        print('\tguard time: ' + str(self.getGuardTime()))

        print('nb of interface bytes: %d' % self.getInterfaceBytesCount())
        print('nb of historical bytes: %d' % self.getHistoricalBytesCount())

    def __str__(self):
        """Returns a string representation of the ATR as a strem of bytes."""
        return toHexString(self.bytes)


if __name__ == '__main__':
    """Small sample illustrating the use of ATR."""

    atrs = [[0x3F, 0x65, 0x25, 0x00, 0x2C, 0x09, 0x69, 0x90, 0x00],
            [0x3F, 0x65, 0x25, 0x08, 0x93, 0x04, 0x6C, 0x90, 0x00],
            [0x3B, 0x16, 0x94, 0x7C, 0x03, 0x01, 0x00, 0x00, 0x0D],
            [0x3B, 0x65, 0x00, 0x00, 0x9C, 0x11, 0x01, 0x01, 0x03],
            [0x3B, 0xE3, 0x00, 0xFF, 0x81, 0x31, 0x52, 0x45, 0xA1,
             0xA2, 0xA3, 0x1B],
            [0x3B, 0xE5, 0x00, 0x00, 0x81, 0x21, 0x45, 0x9C, 0x10,
             0x01, 0x00, 0x80, 0x0D]]

    for atr in atrs:
        a = ATR(atr)
        print(80 * '-')
        print(a)
        a.dump()
        print(toHexString(a.getHistoricalBytes()))



"""Status Word (SW) Exceptions

This module defines the exceptions raised by status word errors or warnings.

__author__ = "http://www.gemalto.com"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""


class SWException(Exception):
    """Base class for status word exceptions.

    Status word exceptions are generated when errors and warnings are detected
    in the sw1 and sw2 bytes of the response apdu.

    """

    def __init__(self, data, sw1, sw2, message=""):
        self.message = message
        """response apdu data"""
        self.data = data
        """response apdu sw1"""
        self.sw1 = sw1
        """response apdu sw2"""
        self.sw2 = sw2

    def __str__(self):
        return repr('Status word exception: ' + self.message + '!')


class WarningProcessingException(SWException):
    """Raised when a warning processing is detected from sw1, sw2.
    Examples of warning processing exception: sw1=62 or sw=63 (ISO7816-4)."""

    def __init__(self, data, sw1, sw2, message=""):
        SWException.__init__(
            self, data, sw1, sw2, "warning processing - " + message)


class ExecutionErrorException(SWException):
    """Raised when an execution error is detected from sw1, sw2.
    Examples of execution error: sw1=64 or sw=65 (ISO7816-4)."""

    def __init__(self, data, sw1, sw2, message=""):
        SWException.__init__(
            self, data, sw1, sw2, "execution error - " + message)


class SecurityRelatedException(SWException):
    """Raised when a security issue is detected from sw1, sw2.
    Examples of security issue: sw1=66 (ISO7816-4)."""

    def __init__(self, data, sw1, sw2, message=""):
        SWException.__init__(
            self, data, sw1, sw2, "security issue - " + message)


class CheckingErrorException(SWException):
    """Raised when a checking error is detected from sw1, sw2.
    Examples of checking error: sw1=67 to 6F (ISO781604)."""

    def __init__(self, data, sw1, sw2, message=""):
        SWException.__init__(
            self, data, sw1, sw2, "checking error - " + message)



#! /usr/bin/env python
"""
Sample for python PCSC wrapper module: send a Control Code to a card or
reader

__author__ = "Ludovic Rousseau"

Copyright 2009-2010 Ludovic Rousseau
Author: Ludovic Rousseau, mailto:ludovic.rousseau@free.fr

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

from __future__ import print_function
from smartcard.scard import *
from smartcard.util import toASCIIBytes
from smartcard.pcsc.PCSCExceptions import *
import sys


def can_do_verify_pin(hCard):
    FEATURE_VERIFY_PIN_DIRECT = 6
    return parse_get_feature_request(hCard, FEATURE_VERIFY_PIN_DIRECT)


def can_do_modify_pin(hCard):
    FEATURE_MODIFY_PIN_DIRECT = 7
    return parse_get_feature_request(hCard, FEATURE_MODIFY_PIN_DIRECT)


def parse_get_feature_request(hCard, feature):
    # check the reader can do a verify pin
    CM_IOCTL_GET_FEATURE_REQUEST = SCARD_CTL_CODE(3400)
    hresult, response = SCardControl(hcard, CM_IOCTL_GET_FEATURE_REQUEST, [])
    if hresult != SCARD_S_SUCCESS:
        raise BaseSCardException(hresult)
    print(response)
    while len(response) > 0:
        tag = response[0]
        if feature == tag:
            return (((((response[2] << 8) +
                    response[3]) << 8) +
                    response[4]) << 8) + response[5]
        response = response[6:]


def verifypin(hCard, control=None):
    if control is None:
        control = can_do_verify_pin(hCard)
        if control is None:
            raise Exception("Not a pinpad")

    command = [0x00,  # bTimerOut
               0x00,  # bTimerOut2
               0x82,  # bmFormatString
               0x04,  # bmPINBlockString
               0x00,  # bmPINLengthFormat
               0x08, 0x04,  # wPINMaxExtraDigit
               0x02,  # bEntryValidationCondition
               0x01,  # bNumberMessage
               0x04, 0x09,  # wLangId
               0x00,  # bMsgIndex
               0x00, 0x00, 0x00,  # bTeoPrologue
               13, 0, 0, 0,  # ulDataLength
               0x00, 0x20, 0x00, 0x00, 0x08, 0x30, 0x30, 0x30, 0x30, 0x30,
               0x30, 0x30, 0x30]  # abData
    hresult, response = SCardControl(hcard, control, command)
    if hresult != SCARD_S_SUCCESS:
        raise BaseSCardException(hresult)
    return hresult, response

try:
    hresult, hcontext = SCardEstablishContext(SCARD_SCOPE_USER)
    if hresult != SCARD_S_SUCCESS:
        raise EstablishContextException(hresult)
    print('Context established!')

    try:
        hresult, readers = SCardListReaders(hcontext, [])
        if hresult != SCARD_S_SUCCESS:
            raise ListReadersException(hresult)
        print('PCSC Readers:', readers)

        if len(readers) < 1:
            raise Exception("No smart card readers")

        for zreader in readers:

            print('Trying to Control reader:', zreader)

            try:
                hresult, hcard, dwActiveProtocol = SCardConnect(
                    hcontext, zreader, SCARD_SHARE_SHARED,
                    SCARD_PROTOCOL_T0 | SCARD_PROTOCOL_T1)
                if hresult != SCARD_S_SUCCESS:
                    raise BaseSCardException(hresult)
                print('Connected with active protocol', dwActiveProtocol)

                try:
                    SELECT = [0x00, 0xA4, 0x04, 0x00, 0x06, 0xA0, 0x00,
                              0x00, 0x00, 0x18, 0xFF]
                    hresult, response = SCardTransmit(
                        hcard,
                        dwActiveProtocol,
                        SELECT)
                    if hresult != SCARD_S_SUCCESS:
                        raise BaseSCardException(hresult)

                    cmd_verify = can_do_verify_pin(hcard)
                    if (cmd_verify):
                        print("can do verify pin: 0x%08X" % cmd_verify)

                    cmd_modify = can_do_modify_pin(hcard)
                    if (cmd_modify):
                        print("can do modify pin: 0x%08X" % cmd_modify)

                    hresult, response = verifypin(hcard, cmd_verify)
                    print('Control:', response)
                finally:
                    hresult = SCardDisconnect(hcard, SCARD_UNPOWER_CARD)
                    if hresult != SCARD_S_SUCCESS:
                        raise BaseSCardException(hresult)
                    print('Disconnected')

            except error as message:
                print(error, message)

    finally:
        hresult = SCardReleaseContext(hcontext)
        if hresult != SCARD_S_SUCCESS:
            raise ReleaseContextException(hresult)
        print('Released context.')

except error as e:
    print(e)

if 'win32' == sys.platform:
    print('press Enter to continue')
    sys.stdin.read(1)



#! /usr/bin/env python
"""
Sample for python PCSC wrapper module: send a Control Code to a card or
reader

__author__ = "Ludovic Rousseau"

Copyright 2007-2010 Ludovic Rousseau
Author: Ludovic Rousseau, mailto:ludovic.rousseau@free.fr

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

from __future__ import print_function
from smartcard.scard import *
from smartcard.util import toBytes

try:
    hresult, hcontext = SCardEstablishContext(SCARD_SCOPE_USER)
    if hresult != SCARD_S_SUCCESS:
        raise error(
            'Failed to establish context: ' + SCardGetErrorMessage(hresult))
    print('Context established!')

    try:
        hresult, readers = SCardListReaders(hcontext, [])
        if hresult != SCARD_S_SUCCESS:
            raise error(
                'Failed to list readers: ' + SCardGetErrorMessage(hresult))
        print('PCSC Readers:', readers)

        if len(readers) < 1:
            raise error('No smart card readers')

        for zreader in readers:

            print('Trying to Control reader:', zreader)

            try:
                hresult, hcard, dwActiveProtocol = SCardConnect(
                    hcontext, zreader, SCARD_SHARE_DIRECT, SCARD_PROTOCOL_T0)
                if hresult != SCARD_S_SUCCESS:
                    raise error(
                        'Unable to connect: ' + SCardGetErrorMessage(hresult))
                print('Connected with active protocol', dwActiveProtocol)

                try:
                    if 'winscard' == resourceManager:
                        # IOCTL_SMARTCARD_GET_ATTRIBUTE = SCARD_CTL_CODE(2)
                        hresult, response = SCardControl(
                            hcard,
                            SCARD_CTL_CODE(2),
                            toBytes("%.8lx" % SCARD_ATTR_VENDOR_NAME))
                        if hresult != SCARD_S_SUCCESS:
                            raise error(
                                'SCardControl failed: ' +
                                SCardGetErrorMessage(hresult))
                        r = ""
                        for i in range(len(response)):
                            r += "%c" % response[i]
                        print('SCARD_ATTR_VENDOR_NAME:', r)
                    elif 'pcsclite' == resourceManager:
                        # get firmware on Gemplus readers
                        hresult, response = SCardControl(
                            hcard,
                            SCARD_CTL_CODE(1),
                            [0x02])
                        if hresult != SCARD_S_SUCCESS:
                            raise error(
                                'SCardControl failed: ' +
                                SCardGetErrorMessage(hresult))
                        r = ""
                        for i in range(len(response)):
                            r += "%c" % response[i]
                        print('Control:', r)
                finally:
                    hresult = SCardDisconnect(hcard, SCARD_UNPOWER_CARD)
                    if hresult != SCARD_S_SUCCESS:
                        raise error(
                            'Failed to disconnect: ' +
                            SCardGetErrorMessage(hresult))
                    print('Disconnected')

            except error as message:
                print(error, message)

    finally:
        hresult = SCardReleaseContext(hcontext)
        if hresult != SCARD_S_SUCCESS:
            raise error(
                'Failed to release context: ' +
                SCardGetErrorMessage(hresult))
        print('Released context.')

except error as e:
    print(e)

import sys
if 'win32' == sys.platform:
    print('press Enter to continue')
    sys.stdin.read(1)



#! /usr/bin/env python
"""
Sample script that defines a custom card connection observer.

__author__ = "http://www.gemalto.com"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""
from __future__ import print_function
from smartcard.CardType import AnyCardType
from smartcard.CardRequest import CardRequest
from smartcard.CardConnectionObserver import CardConnectionObserver
from smartcard.util import toHexString


class TracerAndSELECTInterpreter(CardConnectionObserver):
    """This observer will interprer SELECT and GET RESPONSE bytes
    and replace them with a human readable string."""

    def update(self, cardconnection, ccevent):

        if 'connect' == ccevent.type:
            print('connecting to ' + cardconnection.getReader())

        elif 'disconnect' == ccevent.type:
            print('disconnecting from ' + cardconnection.getReader())

        elif 'command' == ccevent.type:
            str = toHexString(ccevent.args[0])
            str = str.replace("A0 A4 00 00 02", "SELECT")
            str = str.replace("A0 C0 00 00", "GET RESPONSE")
            print('>', str)

        elif 'response' == ccevent.type:
            if [] == ccevent.args[0]:
                print('<  []', "%-2X %-2X" % tuple(ccevent.args[-2:]))
            else:
                print('<',
                      toHexString(ccevent.args[0]),
                      "%-2X %-2X" % tuple(ccevent.args[-2:]))


# define the apdus used in this script
GET_RESPONSE = [0XA0, 0XC0, 00, 00]
SELECT = [0xA0, 0xA4, 0x00, 0x00, 0x02]
DF_TELECOM = [0x7F, 0x10]


# we request any type and wait for 10s for card insertion
cardtype = AnyCardType()
cardrequest = CardRequest(timeout=10, cardType=cardtype)
cardservice = cardrequest.waitforcard()

# create an instance of our observer and attach to the connection
observer = TracerAndSELECTInterpreter()
cardservice.connection.addObserver(observer)


# connect and send APDUs
# the observer will trace on the console
cardservice.connection.connect()

apdu = SELECT + DF_TELECOM
response, sw1, sw2 = cardservice.connection.transmit(apdu)
if sw1 == 0x9F:
    apdu = GET_RESPONSE + [sw2]
    response, sw1, sw2 = cardservice.connection.transmit(apdu)
else:
    print('no DF_TELECOM')

import sys
if 'win32' == sys.platform:
    print('press Enter to continue')
    sys.stdin.read(1)



from smartcard.util import toHexString
expectedReaders = ['Gemalto PC Twin Reader 00 00']
expectedATRs = [[63, 47, 0, 128, 105, 175, 2, 4, 1, 49, 0, 0, 0, 14, 131, 62, 159, 22]]
expectedATRinReader = {}
for i in range(len(expectedReaders)):
    expectedATRinReader[expectedReaders[i]] = expectedATRs[i]
expectedReaderForATR = {}
for i in range(len(expectedReaders)):
    expectedReaderForATR[toHexString(expectedATRs[i])] = expectedReaders[i]
expectedReaderGroups = ['SCard$DefaultReaders']



"""ReaderGroups manages smart card reader in groups.

__author__ = "gemalto http://www.gemalto.com"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

from __future__ import print_function
from smartcard.Exceptions import SmartcardException
from smartcard.ulist import ulist


class BadReaderGroupException(SmartcardException):
    """Raised when trying to add an invalid reader group."""

    def __init__(self):
        SmartcardException.__init__(self, 'Invalid reader group')


class DeleteSCardDefaultReaderGroupException(SmartcardException):
    """Raised when trying to delete SCard$DefaultReaders reader group."""

    def __init__(self):
        SmartcardException.__init__(
            self, 'SCard$DefaultReaders cannot be deleted')


class innerreadergroups(ulist):
    """Smartcard readers groups private class.

    The readergroups singleton manages the creation of the unique
    instance of this class.
    """

    def __init__(self, initlist=None):
        """Retrieve and store list of reader groups"""
        if None == initlist:
            initlist = self.getreadergroups()
        if None != initlist:
            ulist.__init__(self, initlist)
        self.unremovablegroups = []

    def __onadditem__(self, item):
        """Called when a reader group is added."""
        self.addreadergroup(item)

    def __onremoveitem__(self, item):
        """Called when a reader group is added."""
        self.removereadergroup(item)

    def __iter__(self):
        return ulist.__iter__(self)

    def next(self):
        return ulist.__next__(self)

    #
    # abstract methods implemented in subclasses
    #

    def getreadergroups(self):
        """Returns the list of smartcard reader groups."""
        return []

    def addreadergroup(self, newgroup):
        """Add a reader group"""
        if not isinstance(newgroup, type("")):
            raise BadReaderGroupException
        self += newgroup

    def removereadergroup(self, group):
        """Remove a reader group"""
        if not isinstance(group, type("")):
            raise BadReaderGroupException
        self.remove(group)

    def addreadertogroup(self, readername, groupname):
        """Add a reader to a reader group"""
        pass

    def removereaderfromgroup(self, readername, groupname):
        """Remove a reader from a reader group"""
        pass


class readergroups(object):
    """ReadersGroups organizes smart card reader as groups."""

    """The single instance of __readergroups"""
    instance = None
    innerclazz = innerreadergroups

    def __init__(self, initlist=None):
        """Create a single instance of innerreadergroups on first call"""
        if None == readergroups.instance:
            readergroups.instance = self.innerclazz(initlist)

    """All operators redirected to inner class."""

    def __getattr__(self, name):
        return getattr(self.instance, name)


if __name__ == '__main__':
    print(readergroups())




#! /usr/bin/env python
"""
Setup script to build a standalone cardmonitor.exe executable on windows
using py2exe. Run: python.exe setup.py py2exe, to build executable file.

__author__ = "http://www.gemalto.com"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

from distutils.core import setup
import py2exe

from smartcard.wx import ICO_SMARTCARD, ICO_READER

Mydata_files = [('images',
                 ['images/mysmartcard.ico',
                  ICO_SMARTCARD, ICO_READER])]

setup(windows=['cardmonitor.py'],
    data_files=Mydata_files,
    options={"py2exe": {"dll_excludes": ["MSVCP90.dll"]}}
)



"""PyroReaderClient: concrete reader class for Remote Readers

__author__ = "gemalto http://www.gemalto.com"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

from __future__ import print_function
import Pyro.core
import Pyro.naming

from smartcard.Exceptions import NoCardException
from smartcard.reader.Reader import Reader


class PyroReader(Reader):
    """Remote reader class."""
    def __init__(self, readername):
        """Constructs a new Remote Reader client implementation from a
        Pyro URI."""
        ns = Pyro.naming.NameServerLocator().getNS()
        self.uri = ns.resolve(':pyscard.smartcard.readers.' + readername)
        self.reader = Pyro.core.getAttrProxyForURI(self.uri)
        self.name = self.reader.name

    def addtoreadergroup(self, groupname):
        """Add reader to a reader group."""
        self.reader.addtoreadergroup(groupname)

    def removefromreadergroup(self, groupname):
        """Remove a reader from a reader group"""
        self.reader.removefromreadergroup(groupname)

    def createConnection(self):
        """Return a card connection thru a remote reader."""
        uri = self.reader.createConnection()
        return Pyro.core.getAttrProxyForURI(uri)

    class Factory:
        def create(readername):
            return PyroReader(readername)
        create = staticmethod(create)

    def readers(groups=[]):
        readernames = []
        try:
            ns = Pyro.naming.NameServerLocator().getNS()
            readernames = ns.list(':pyscard.smartcard.readers')
        except Pyro.errors.NamingError:
            print('Warning: pyro name server not found')

        remotereaders = []
        for readername in readernames:
            remotereaders.append(PyroReader.Factory.create(readername[0]))

        return remotereaders
    readers = staticmethod(readers)

if __name__ == '__main__':
    SELECT = [0xA0, 0xA4, 0x00, 0x00, 0x02]
    DF_TELECOM = [0x7F, 0x10]
    from smartcard.util import *

    remotereaders = PyroReader.readers()
    for reader in remotereaders:
        try:
            print(reader.name, ', uri: ', reader.uri)
            connection = reader.createConnection()
            connection.connect()
            print(toHexString(connection.getATR()))
            data, sw1, sw2 = connection.transmit(SELECT + DF_TELECOM)
            print("%X %X" % (sw1, sw2))
        except NoCardException as x:
            print('no card in reader')



#! /usr/bin/env python
"""Setup file for distutils

__author__ = "http://www.gemalto.com"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

Copyright 2007-2018 Ludovic Rousseau ludovic.rousseau@free.fr

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

from distutils.util import get_platform
import sys

from setuptools import setup, Extension
from setuptools.command.build_py import build_py


if sys.version_info[0:2] < (2, 6):
    raise RuntimeError("pyscard requires Python 2.6+ to build.")

if get_platform() in ('win32', 'win-amd64'):
    platform__cc_defines = [('WIN32', '100')]
    platform_swig_opts = ['-DWIN32']
    platform_sources = ['smartcard/scard/scard.rc']
    platform_libraries = ['winscard']
    platform_include_dirs = []
    platform_extra_compile_args = []
    platform_extra_link_args = []

elif 'cygwin-' in get_platform():
    platform__cc_defines = [('WIN32', '100')]
    platform_swig_opts = ['-DWIN32']
    platform_sources = []
    platform_libraries = ['winscard']
    platform_include_dirs = []
    platform_extra_compile_args = []
    platform_extra_link_args = []

elif 'macosx-10.' in get_platform():
    if 'macosx-10.6' in get_platform():
        macosx_define = '__LEOPARD__'  # Snow Leopard, Python 2.6
    else:
        macosx_define = '__LION__'  # Lion (and above), Python 2.7
    platform__cc_defines = [('PCSCLITE', '1'),
                            ('__APPLE__', '1'),
                            (macosx_define, '1')]
    platform_swig_opts = ['-DPCSCLITE', '-D__APPLE__', '-D' + macosx_define]
    platform_sources = []
    platform_libraries = []
    platform_include_dirs = []
    platform_extra_compile_args = ['-v', '-arch', 'i386',
                                   '-arch', 'x86_64', '-ggdb']
    platform_extra_link_args = ['-arch', 'i386', '-arch', 'x86_64', '-ggdb']

# Other (GNU/Linux, etc.)
#
else:
    platform__cc_defines = [('PCSCLITE', '1')]
    platform_swig_opts = ['-DPCSCLITE']
    platform_sources = []
    platform_libraries = []
    platform_include_dirs = ['/usr/include/PCSC']
    platform_extra_compile_args = []    # ['-ggdb', '-O0']
    platform_extra_link_args = []   # ['-ggdb']


VERSION_INFO = (1, 9, 7, 0)
VERSION_STR = '%i.%i.%i' % VERSION_INFO[:3]
VERSION_ALT = '%i,%01i,%01i,%04i' % VERSION_INFO


class BuildPyBuildExtFirst(build_py):
    """Workaround substitude `build_py` command for SWIG"""
    def run(self):
        # Run build_ext first so that SWIG generated files are included
        self.run_command('build_ext')
        return build_py.run(self)


kw = {'name': "pyscard",
      'version': VERSION_STR,
      'description': "Smartcard module for Python.",
      'author': "Ludovic Rousseau",
      'author_email': "ludovic.rousseau@free.fr",
      'url': "https://github.com/LudovicRousseau/pyscard",
      'long_description': 'Smartcard package for Python',
      'platforms': ['linux', 'win32'],
      'packages': ["smartcard",
                   "smartcard.pcsc",
                   "smartcard.pyro",
                   "smartcard.reader",
                   "smartcard.scard",
                   "smartcard.sw",
                   "smartcard.util",
                   "smartcard.wx",
                   ],
      'package_dir': {"": "."},
      'package_data': {
                         "smartcard.wx": ["resources/*.ico"],
                       },

      'cmdclass': {'build_py': BuildPyBuildExtFirst},
      # the _scard.pyd extension to build
      'ext_modules': [Extension("smartcard.scard._scard",
                      define_macros=[
                                    ('VER_PRODUCTVERSION', VERSION_ALT),
                                    ('VER_PRODUCTVERSION_STR', VERSION_STR)] \
                      + platform__cc_defines,
                      include_dirs=['smartcard/scard/'] \
                      + platform_include_dirs,
                      sources=["smartcard/scard/helpers.c",
                               "smartcard/scard/winscarddll.c",
                               "smartcard/scard/scard.i"] \
                                       + platform_sources,
                      libraries=platform_libraries,
                      extra_compile_args=platform_extra_compile_args,
                      extra_link_args=platform_extra_link_args,
                      swig_opts=['-outdir',
                                 'smartcard/scard'] \
                      + platform_swig_opts)],
      'extras_require': {
            'Gui': ['wxPython'],
            'Pyro': ['Pyro'],
            },

      'test_suite': 'test',

      'classifiers': [
          'Development Status :: 5 - Production/Stable',
          'License :: OSI Approved :: GNU Lesser General Public License v2 '
          'or later (LGPLv2+)',
          'Intended Audience :: Developers',
          'Operating System :: Unix',
          'Operating System :: Microsoft :: Windows',
          'Operating System :: MacOS :: MacOS X',
          'Programming Language :: Python :: 2.6',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3.3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Topic :: Security',
          ]
      }

# FIXME Sourceforge downloads are unauthenticated, migrate to PyPI
kw['download_url'] = ('http://sourceforge.net/projects/%(name)s/files'
                      '/%(name)s/%(name)s%%20%(version)s'
                      '/%(name)s-%(version)s.tar.gz/download' % kw)

setup(**kw)



#! /usr/bin/env python
"""
Sample script that monitors card insertions,
connects to cards and transmit an apdu

__author__ = "http://www.gemalto.com"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""
from __future__ import print_function
from time import sleep

from smartcard.CardMonitoring import CardMonitor, CardObserver
from smartcard.util import *

# replace by your favourite apdu
SELECT_DF_TELECOM = [0xA0, 0xA4, 0x00, 0x00, 0x02, 0x7F, 0x10]


class transmitobserver(CardObserver):
    """A card observer that is notified when cards are inserted/removed
    from the system, connects to cards and SELECT DF_TELECOM """

    def __init__(self):
        self.cards = []

    def update(self, observable, actions):
        (addedcards, removedcards) = actions
        for card in addedcards:
            if card not in self.cards:
                self.cards += [card]
                print("+Inserted: ", toHexString(card.atr))
                card.connection = card.createConnection()
                card.connection.connect()
                response, sw1, sw2 = card.connection.transmit(
                    SELECT_DF_TELECOM)
                print("%.2x %.2x" % (sw1, sw2))

        for card in removedcards:
            print("-Removed: ", toHexString(card.atr))
            if card in self.cards:
                self.cards.remove(card)

if __name__ == '__main__':
    print("Insert or remove a smartcard in the system.")
    print("This program will exit in 100 seconds")
    print("")
    cardmonitor = CardMonitor()
    cardobserver = transmitobserver()
    cardmonitor.addObserver(cardobserver)

    sleep(100)



"""Graphical APDU Tracer.

__author__ = "gemalto http://www.gemalto.com"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

# wxPython GUI modules (http://www.wxpython.org)
import wx

from smartcard.CardConnectionObserver import CardConnectionObserver
from smartcard.util import toHexString

[
wxID_APDUTEXTCTRL,
] = [wx.NewId() for x in range(1)]


class APDUTracerPanel(wx.Panel, CardConnectionObserver):

    def __init__(self, parent):
        wx.Panel.__init__(self, parent, -1)

        boxsizer = wx.BoxSizer(wx.HORIZONTAL)
        self.apdutextctrl = wx.TextCtrl(
            self, wxID_APDUTEXTCTRL,
            "",
            pos=wx.DefaultPosition,
            style=wx.TE_MULTILINE | wx.TE_READONLY)
        boxsizer.Add(self.apdutextctrl, 1, wx.EXPAND | wx.ALL, 5)
        self.SetSizer(boxsizer)
        self.SetAutoLayout(True)

        self.Bind(wx.EVT_TEXT_MAXLEN, self.OnMaxLength, self.apdutextctrl)

    def OnMaxLength(self, evt):
        '''Reset text buffer when max length is reached.'''
        self.apdutextctrl.SetValue("")
        evt.Skip()

    def update(self, cardconnection, ccevent):
        '''CardConnectionObserver callback.'''

        apduline = ""
        if 'connect' == ccevent.type:
            apduline += 'connecting to ' + cardconnection.getReader()

        elif 'disconnect' == ccevent.type:
            apduline += 'disconnecting from ' + cardconnection.getReader()

        elif 'command' == ccevent.type:
            apduline += '> ' + toHexString(ccevent.args[0])

        elif 'response' == ccevent.type:
            if [] == ccevent.args[0]:
                apduline += "< %-2X %-2X" % tuple(ccevent.args[-2:])
            else:
                apduline += "< " + toHexString(ccevent.args[0]) + \
                            "%-2X %-2X" % tuple(ccevent.args[-2:])

        self.apdutextctrl.AppendText(apduline + "\n")



#! /usr/bin/env python
"""
Sample for python PCSC wrapper module: illustrate reader groups functions

__author__ = "http://www.gemalto.com"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com
Copyright 2010 Ludovic Rousseau
Author: Ludovic Rousseau, mailto:ludovic.rousseau@free.fr

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

from __future__ import print_function
from smartcard.scard import *

newgroup = 'MyReaderGroup'

try:
    hresult, hcontext = SCardEstablishContext(SCARD_SCOPE_USER)
    if hresult != SCARD_S_SUCCESS:
        raise error(
            'Failed to establish context: ' +
            SCardGetErrorMessage(hresult))
    print('Context established!')

    try:
        hresult, readers = SCardListReaders(hcontext, [])
        if hresult != SCARD_S_SUCCESS:
            raise error(
                'Failed to list readers: ' +
                SCardGetErrorMessage(hresult))
        print('PCSC Readers in all groups:', readers)

        hresult, readerGroups = SCardListReaderGroups(hcontext)
        if hresult != SCARD_S_SUCCESS:
            raise error(
                'Unable to list reader groups: ' +
                SCardGetErrorMessage(hresult))
        print('PCSC Reader groups:', readerGroups)

        if 'winscard' == resourceManager:

            hresult = SCardIntroduceReaderGroup(hcontext, newgroup)
            if hresult != SCARD_S_SUCCESS:
                raise error(
                    'Unable to introduce reader group: ' +
                    SCardGetErrorMessage(hresult))

            dummyreader = readers[0] + ' dummy'
            hresult = SCardIntroduceReader(hcontext, dummyreader, readers[0])
            if hresult != SCARD_S_SUCCESS:
                raise error(
                    'Unable to introduce reader: ' +
                    dummyreader + ' : ' +
                    SCardGetErrorMessage(hresult))

            hresult, readers = SCardListReaders(hcontext, [])
            if hresult != SCARD_S_SUCCESS:
                raise error(
                    'Failed to list readers: ' +
                    SCardGetErrorMessage(hresult))
            print('PCSC Readers in all groups:', readers)

            hresult = SCardAddReaderToGroup(hcontext, dummyreader, newgroup)
            if hresult != SCARD_S_SUCCESS:
                raise error(
                    'Unable to add reader to group: ' +
                    SCardGetErrorMessage(hresult))

            hresult, readerGroups = SCardListReaderGroups(hcontext)
            if hresult != SCARD_S_SUCCESS:
                raise error(
                    'Unable to list reader groups: ' +
                    SCardGetErrorMessage(hresult))
            print('PCSC Reader groups:', readerGroups)

            hresult, readers = SCardListReaders(hcontext, [newgroup])
            if hresult != SCARD_S_SUCCESS:
                raise error(
                    'Failed to list readers in group ' +
                    newgroup + ' : ' +
                    SCardGetErrorMessage(hresult))
            print('PCSC Readers in reader group', newgroup, ':', readers)

            hresult = SCardRemoveReaderFromGroup(
                hcontext,
                dummyreader,
                newgroup)
            if hresult != SCARD_S_SUCCESS:
                raise error(
                    'Unable to remove reader from group: ' +
                    SCardGetErrorMessage(hresult))

            hresult, readerGroups = SCardListReaderGroups(hcontext)
            if hresult != SCARD_S_SUCCESS:
                raise error(
                    'Unable to list reader groups: ' +
                    SCardGetErrorMessage(hresult))
            print('PCSC Reader groups:', readerGroups)

            hresult = SCardForgetReaderGroup(hcontext, newgroup)
            if hresult != SCARD_S_SUCCESS:
                raise error(
                    'Unable to forget reader group: ' +
                    SCardGetErrorMessage(hresult))

            hresult = SCardForgetReader(hcontext, dummyreader)
            if hresult != SCARD_S_SUCCESS:
                raise error(
                    'Failed to forget readers ' +
                    SCardGetErrorMessage(hresult))

            hresult, readers = SCardListReaders(hcontext, [])
            if hresult != SCARD_S_SUCCESS:
                raise error(
                    'Failed to list readers: ' +
                    SCardGetErrorMessage(hresult))
            print('PCSC Readers in all groups:', readers)

        elif 'pcsclite' == resourceManager:
            hresult, readers = SCardListReaders(hcontext, readerGroups)
            if hresult != SCARD_S_SUCCESS:
                raise error(
                    'Failed to list readers in groups ' +
                    repr(readerGroups) + ' : ' +
                    SCardGetErrorMessage(hresult))
            print('PCSC Readers in reader group', readerGroups, ':', readers)

    finally:
        hresult = SCardReleaseContext(hcontext)
        if hresult != SCARD_S_SUCCESS:
            raise error(
                'Failed to release context: ' +
                SCardGetErrorMessage(hresult))
        print('Released context.')

    import sys
    if 'win32' == sys.platform:
        print('press Enter to continue')
        sys.stdin.read(1)

except error as e:
    print(e)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

# to execute:
# $ cd test
# $ python -m unittest

import unittest
from smartcard.scard import SCardGetErrorMessage
from smartcard.scard import SCARD_S_SUCCESS, SCARD_F_INTERNAL_ERROR
from distutils.util import get_platform


class TestError(unittest.TestCase):

    def test_SCardGetErrorMessage(self):
        res = SCardGetErrorMessage(SCARD_S_SUCCESS)

        # do not test on Windows
        # the error messages are different and localized
        if get_platform() in ('win32', 'win-amd64'):
            return

        expected = "Command successful."
        self.assertEqual(res, expected)

        res = SCardGetErrorMessage(SCARD_F_INTERNAL_ERROR)
        expected = "Internal error."
        self.assertEqual(res, expected)

        res = SCardGetErrorMessage(1)
        # macOS bug not yet fixed
        if get_platform().startswith('macosx-'):
            expected = "Unkown error: 0x00000001"
        else:
            expected = "Unknown error: 0x00000001"
        self.assertEqual(res, expected)


if __name__ == '__main__':
    unittest.main()
